@@ -9,25 +9,52 @@ from torch.utils.tensorboard import SummaryWriter
 from torch.optim.lr_scheduler import (
     CosineAnnealingLR, CosineAnnealingWarmRestarts, StepLR, CyclicLR)
 import pytorch_lightning as pl
-from deepblast.alignment import NeedlemanWunschAligner
-from deepblast.dataset.alphabet import UniprotTokenizer
+from deepblast.alignment import NeuralAligner
 from deepblast.dataset import TMAlignDataset
 from deepblast.dataset.utils import (
     decode, states2edges, collate_f, test_collate_f,
-    unpack_sequences, pack_sequences, revstate_f)
+    unpack_sequences, pack_sequences, revstate_f,
+    get_sequence
+)
 from deepblast.losses import (
     SoftAlignmentLoss, SoftPathLoss, MatrixCrossEntropy)
 from deepblast.score import (roc_edges, alignment_visualization,
-                             alignment_text)
+                             alignment_text, filter_gaps)
 
 
-class LightningAligner(pl.LightningModule):
+class DeepBLAST(pl.LightningModule):
+
+    def __init__(self, batch_size=20,
+                 hidden_dim=1024,
+                 embedding_dim=1024,
+                 epochs=32,
+                 finetune=False,
+                 layers=1,
+                 dropout=0,
+                 lm=None,
+                 tokenizer=None,
+                 learning_rate=0.0001,
+                 loss='cross_entropy',
+                 mask_gaps=False,
+                 multitask=False,
+                 num_workers=1,
+                 output_directory=None,
+                 scheduler='cosine',
+                 test_pairs=None,
+                 train_pairs=None,
+                 valid_pairs=None,
+                 visualization_fraction=1.0,
+                 shuffle_validation=False,
+                 device='gpu',
+                 alignment_mode='needleman-wunsch'
+    ):
+
+        super(DeepBLAST, self).__init__()
+        self.save_hyperparameters(ignore=['lm', 'tokenizer'])
+        assert tokenizer is not None
+        assert lm is not None
+        self.tokenizer = tokenizer
 
-    def __init__(self, args):
-        super(LightningAligner, self).__init__()
-        self.tokenizer = UniprotTokenizer(pad_ends=False)
-        self.hparams = args
-        self.initialize_aligner()
         if self.hparams.loss == 'sse':
             self.loss_func = SoftAlignmentLoss()
         elif self.hparams.loss == 'cross_entropy':
@@ -35,22 +62,23 @@ class LightningAligner(pl.LightningModule):
         elif self.hparams.loss == 'path':
             self.loss_func = SoftPathLoss()
         else:
-            raise ValueError(f'`{args.loss}` is not implemented.')
+            raise ValueError(f'`{self.hparams.loss}` is not implemented.')
 
-    def initialize_aligner(self):
-        n_alpha = len(self.tokenizer.alphabet)
+        # initialize_aligner
         n_embed = self.hparams.embedding_dim
-        n_input = self.hparams.rnn_input_dim
-        n_units = self.hparams.rnn_dim
+        n_input = self.hparams.hidden_dim
+        n_units = self.hparams.hidden_dim
         n_layers = self.hparams.layers
-        self.aligner = NeedlemanWunschAligner(
-            n_alpha, n_input, n_units, n_embed, n_layers)
+        dropout = self.hparams.dropout
+        self.aligner = NeuralAligner(
+            n_input, n_units, n_embed, n_layers, dropout=dropout, lm=lm,
+            alignment_mode=alignment_mode,
+            device=device)
+
 
     def align(self, x, y):
-        x_code = torch.Tensor(self.tokenizer(str.encode(x))).long()
-        y_code = torch.Tensor(self.tokenizer(str.encode(y))).long()
-        x_code = x_code.to(self.device)
-        y_code = y_code.to(self.device)
+        x_code = get_sequence(x, self.tokenizer)[0].to(self.device)
+        y_code = get_sequence(y, self.tokenizer)[0].to(self.device)
         seq, order = pack_sequences([x_code], [y_code])
         gen = self.aligner.traceback(seq, order)
         decoded, _ = next(gen)
@@ -91,6 +119,7 @@ class LightningAligner(pl.LightningModule):
     def train_dataloader(self):
         train_dataset = TMAlignDataset(
             self.hparams.train_pairs,
+            tokenizer=self.tokenizer,
             construct_paths=isinstance(self.loss_func, SoftPathLoss))
         train_dataloader = DataLoader(
             train_dataset, self.hparams.batch_size, collate_fn=collate_f,
@@ -101,16 +130,19 @@ class LightningAligner(pl.LightningModule):
     def val_dataloader(self):
         valid_dataset = TMAlignDataset(
             self.hparams.valid_pairs,
+            tokenizer=self.tokenizer,
             construct_paths=isinstance(self.loss_func, SoftPathLoss))
         valid_dataloader = DataLoader(
             valid_dataset, self.hparams.batch_size, collate_fn=collate_f,
-            shuffle=False, num_workers=self.hparams.num_workers,
+            shuffle=self.hparams.shuffle_validation,
+            num_workers=self.hparams.num_workers,
             pin_memory=True)
         return valid_dataloader
 
     def test_dataloader(self):
         test_dataset = TMAlignDataset(
             self.hparams.test_pairs, return_names=True,
+            tokenizer=self.tokenizer,
             construct_paths=isinstance(self.loss_func, SoftPathLoss))
         test_dataloader = DataLoader(
             test_dataset, self.hparams.batch_size, shuffle=False,
@@ -139,18 +171,18 @@ class LightningAligner(pl.LightningModule):
 
     def training_step(self, batch, batch_idx):
         self.aligner.train()
-        genes, others, s, A, P, G = batch
+        genes, others, s, A, P, G, _, _ = batch
         seq, order = pack_sequences(genes, others)
         predA, theta, gap = self.aligner(seq, order)
         _, xlen, _, ylen = unpack_sequences(seq, order)
         loss = self.compute_loss(xlen, ylen, predA, A, P, G, theta)
         assert torch.isnan(loss).item() is False
-        if len(self.trainer.lr_schedulers) >= 1:
-            current_lr = self.trainer.lr_schedulers[0]['scheduler']
-            current_lr = current_lr.get_last_lr()[0]
-        else:
-            current_lr = self.hparams.learning_rate
-        tensorboard_logs = {'train_loss': loss, 'lr': current_lr}
+        # if len(self.trainer.lr_schedulers) >= 1:
+        #     current_lr = self.trainer.lr_schedulers[0]['scheduler']
+        #     current_lr = current_lr.get_last_lr()[0]
+        # else:
+        #     current_lr = self.hparams.learning_rate
+        tensorboard_logs = {'train_loss': loss}
         # log the learning rate
         return {'loss': loss, 'log': tensorboard_logs}
 
@@ -161,16 +193,18 @@ class LightningAligner(pl.LightningModule):
             # TODO: Issue #47
             x_str = decode(
                 list(x[b, :xlen[b]].squeeze().cpu().detach().numpy()),
-                self.tokenizer.alphabet)
+                self.tokenizer.get_vocab())
             y_str = decode(
                 list(y[b, :ylen[b]].squeeze().cpu().detach().numpy()),
-                self.tokenizer.alphabet)
+                self.tokenizer.get_vocab())
             decoded, _ = next(gen)
             pred_x, pred_y, pred_states = list(zip(*decoded))
             pred_states = np.array(list(pred_states))
-            truth_states = states[b].cpu().detach().numpy()
+            true_states = states[b].cpu().detach().numpy()
             pred_edges = states2edges(pred_states)
-            true_edges = states2edges(truth_states)
+            true_edges = states2edges(true_states)
+            pred_edges = filter_gaps(pred_states, pred_edges)
+            true_edges = filter_gaps(true_states, true_edges)
             stats = roc_edges(true_edges, pred_edges)
             if random.random() < self.hparams.visualization_fraction:
                 Av = A[b].cpu().detach().numpy().squeeze()
@@ -184,39 +218,46 @@ class LightningAligner(pl.LightningModule):
                     self.global_step, close=True)
                 try:
                     text = alignment_text(
-                        x_str, y_str, pred_states, truth_states, stats)
+                        x_str, y_str, pred_states, true_states, stats)
                     self.logger.experiment.add_text(
                         f'alignment/{batch_idx}/{b}', text, self.global_step)
                 except Exception as e:
-                    print(predA[b])
-                    print(A[b])
-                    print(theta[b])
-                    print(xlen[b], ylen[b])
-                    raise e
+                    print('predA', predA[b].shape)
+                    print('A', A[b].shape)
+                    print('theta', theta[b].shape)
+                    print('xlen', xlen[b], 'ylen', ylen[b])
+                    print('pred_states', pred_states, len(pred_states))
+                    print('true_states', true_states, len(true_states))
             statistics.append(stats)
         return statistics
 
     def validation_step(self, batch, batch_idx):
-        genes, others, s, A, P, G = batch
+        genes, others, s, A, P, G, _, _ = batch
         seq, order = pack_sequences(genes, others)
         predA, theta, gap = self.aligner(seq, order)
         x, xlen, y, ylen = unpack_sequences(seq, order)
         loss = self.compute_loss(xlen, ylen, predA, A, P, G, theta)
+
         assert torch.isnan(loss).item() is False
-        # Obtain alignment statistics + visualizations
-        gen = self.aligner.traceback(seq, order)
-        # TODO; compare the traceback and the forward
-        statistics = self.validation_stats(
-            x, y, xlen, ylen, gen, s, A, predA, theta, gap, batch_idx)
-        statistics = pd.DataFrame(
-            statistics, columns=[
-                'val_tp', 'val_fp', 'val_fn', 'val_perc_id',
-                'val_ppv', 'val_fnr', 'val_fdr'
-            ]
-        )
-        statistics = statistics.mean(axis=0).to_dict()
+
+        self.log('validation_loss', loss,
+                 batch_size=self.hparams.batch_size, sync_dist=True)
         tensorboard_logs = {'valid_loss': loss}
-        tensorboard_logs = {**tensorboard_logs, **statistics}
+
+        # Obtain alignment statistics + visualizations
+        if self.hparams.visualization_fraction > 0:
+            gen = self.aligner.traceback(seq, order)
+            # TODO; compare the traceback and the forward
+            statistics = self.validation_stats(
+                x, y, xlen, ylen, gen, s, A, predA, theta, gap, batch_idx)
+            statistics = pd.DataFrame(
+                statistics, columns=[
+                    'val_tp', 'val_fp', 'val_fn', 'val_perc_id',
+                    'val_ppv', 'val_fnr', 'val_fdr'
+                ]
+            )
+            statistics = statistics.mean(axis=0).to_dict()
+            tensorboard_logs = {**tensorboard_logs, **statistics}
         return {'validation_loss': loss,
                 'log': tensorboard_logs}
 
@@ -256,24 +297,29 @@ class LightningAligner(pl.LightningModule):
         losses = list(map(loss_f, outputs))
         loss = sum(losses) / len(losses)
         self.logger.experiment.add_scalar('val_loss', loss, self.global_step)
-        metrics = ['val_tp', 'val_fp', 'val_fn', 'val_perc_id',
-                   'val_ppv', 'val_fnr', 'val_fdr']
-        scores = []
-        for i, m in enumerate(metrics):
-            loss_f = lambda x: x['log'][m]
-            losses = list(map(loss_f, outputs))
-            scalar = sum(losses) / len(losses)
-            scores.append(scalar)
-            self.logger.experiment.add_scalar(m, scalar, self.global_step)
+        # self.log('validation_loss') = loss
+
+        # metrics = ['val_tp', 'val_fp', 'val_fn', 'val_perc_id',
+        #            'val_ppv', 'val_fnr', 'val_fdr']
+        # scores = []
+        # for i, m in enumerate(metrics):
+        #     loss_f = lambda x: x['log'][m]
+        #     losses = list(map(loss_f, outputs))
+        #     scalar = sum(losses) / len(losses)
+        #     scores.append(scalar)
+        #     self.logger.experiment.add_scalar(m, scalar, self.global_step)
 
         tensorboard_logs = dict(
-            [('val_loss', loss)] + list(zip(metrics, scores))
+            [('val_loss', loss)] # + list(zip(metrics, scores))
         )
         return {'val_loss': loss, 'log': tensorboard_logs}
 
     def configure_optimizers(self):
-        for p in self.aligner.lm.parameters():
-            p.requires_grad = False
+        # Freeze language model
+        if self.hparams.finetune is False:
+            for param in (self.aligner.lm.parameters()):
+                param.requires_grad = False
+
         grad_params = list(filter(
             lambda p: p.requires_grad, self.aligner.parameters()))
         optimizer = torch.optim.AdamW(
@@ -314,17 +360,22 @@ class LightningAligner(pl.LightningModule):
         parser.add_argument(
             '--valid-pairs', help='Validation pairs file', required=True)
         parser.add_argument(
-            '--embedding-dim', help='Embedding dimension (default 512).',
-            required=False, type=int, default=512)
+            '--pretrain-path', help='Path to pretrained protein language model',
+            required=False, type=str, default='Rostlab/prot_t5_xl_uniref50')
         parser.add_argument(
-            '--rnn-input-dim', help='RNN input dimension (default 512).',
-            required=False, type=int, default=512)
+            '--embedding-dim',
+            help='Embedding dimension for aligner (default 1024).',
+            required=False, type=int, default=1024)
         parser.add_argument(
-            '--rnn-dim', help='Number of hidden RNN units (default 512).',
-            required=False, type=int, default=512)
+            '--hidden-dim',
+            help='Hidden dimension for intermediate layers (default 1024).',
+            required=False, type=int, default=1024)
         parser.add_argument(
             '--layers', help='Number of RNN layers (default 2).',
             required=False, type=int, default=2)
+        parser.add_argument(
+            '--dropout', help='Dropout rate.',
+            required=False, type=float, default=0.5)
         parser.add_argument(
             '--loss',
             help=('Loss function. Options include {sse, path, cross_entropy} '
@@ -345,10 +396,16 @@ class LightningAligner(pl.LightningModule):
                 'WARNING: this option is deprecated, use at your own risk.'
             )
         )
+        parser.add_argument(
+            '--mode', default=False, required=False, type=str,
+            help=(
+                'Aligment mode. Choices include `needleman-wunch`, '
+                '`smith-waterman`, or `gotoh`.'
+            )
+        )
         parser.add_argument(
             '--finetune',
-            help=('Perform finetuning. '
-                  'WARNING: this option is not tested, use at your own risk.'),
+            help=('Perform finetuning of language model.'),
             default=False, required=False, type=bool)
         parser.add_argument(
             '--mask-gaps',
@@ -367,6 +424,10 @@ class LightningAligner(pl.LightningModule):
             '--visualization-fraction',
             help='Fraction of alignments to be visualized per epoch',
             required=False, type=float, default=0.1)
+        parser.add_argument(
+            '--shuffle-validation',
+            help='Whether or not to shuffle the validation data.',
+            required=False, type=bool, default=False)
         parser.add_argument(
             '-o', '--output-directory',
             help='Output directory of model results', required=True)
